#include <stdio.h>
#include "chem_ext.h"
#include "chem_g@.h"

void STARTUP_chem() {
  GenesisObject* object;
  GenesisObject  tobject;
  char*          slotnames[50];
  char*          argv[50];

  /* Definition of object diffusion */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "diffusion";
  tobject.type = "diffusion_type";
  tobject.size = sizeof(struct diffusion_type);
  { extern int diffusion(); tobject.function = diffusion; HashFunc("diffusion", diffusion, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("channel"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("diffusion");
  object->defaults->object = object;
  object->defaults->name = CopyString("diffusion");
  object->author = "A. Blackwell GMU Dec 97; revised GMU Dec 99; GMU Mar 02";
  { extern int diffusion(); AddActionToObject(object, "INIT", diffusion, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'diffusion'\n")); HashFunc("diffusion", diffusion, "int"); }
  { extern int diffusion(); AddActionToObject(object, "PROCESS", diffusion, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'diffusion'\n")); HashFunc("diffusion", diffusion, "int"); }
  { extern int diffusion(); AddActionToObject(object, "RESET", diffusion, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'diffusion'\n")); HashFunc("diffusion", diffusion, "int"); }
  { extern int diffusion(); AddActionToObject(object, "CHECK", diffusion, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'diffusion'\n")); HashFunc("diffusion", diffusion, "int"); }
  slotnames[0] = "len1";
  slotnames[1] = "area1";
  slotnames[2] = "conc1";
  MsgListAdd(object, "POOL1", diffusion_POOL1, slotnames, 3);
  slotnames[0] = "len2";
  slotnames[1] = "area2";
  slotnames[2] = "conc2";
  MsgListAdd(object, "POOL2", diffusion_POOL2, slotnames, 3);
  object->description = "Computes flux in molecules due to diffusion through surface between two rxnpools\n(1-D or 2-D). Input must be concentration (not quantity). return messages:\n addmsg from diffusion object to pool2 RXN0MOLES difflux2 \n addmsg from diffusion object to pool1 RXN0MOLES difflux1 \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object rxnpool */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "rxnpool";
  tobject.type = "rxnpool_type";
  tobject.size = sizeof(struct rxnpool_type);
  { extern int rxnpool(); tobject.function = rxnpool; HashFunc("rxnpool", rxnpool, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("concentration"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("rxnpool");
  object->defaults->object = object;
  object->defaults->name = CopyString("rxnpool");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99; GMU Mar 02";
  { extern int rxnpool(); AddActionToObject(object, "INIT", rxnpool, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  { extern int rxnpool(); AddActionToObject(object, "PROCESS", rxnpool, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  { extern int rxnpool(); AddActionToObject(object, "RESET", rxnpool, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  { extern int rxnpool(); AddActionToObject(object, "CHECK", rxnpool, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  { extern int rxnpool(); AddActionToObject(object, "SAVE2", rxnpool, 0) ? 0 : (Error(), printf("adding action 'SAVE2' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  { extern int rxnpool(); AddActionToObject(object, "RESTORE2", rxnpool, 0) ? 0 : (Error(), printf("adding action 'RESTORE2' to object 'rxnpool'\n")); HashFunc("rxnpool", rxnpool, "int"); }
  slotnames[0] = "A";
  MsgListAdd(object, "RXN0", rxnpool_RXN0, slotnames, 1);
  slotnames[0] = "B";
  MsgListAdd(object, "RXN1", rxnpool_RXN1, slotnames, 1);
  slotnames[0] = "A";
  slotnames[1] = "B";
  MsgListAdd(object, "RXN2", rxnpool_RXN2, slotnames, 2);
  slotnames[0] = "A";
  slotnames[1] = "B";
  MsgListAdd(object, "RXN2MOLES", rxnpool_RXN2MOLES, slotnames, 2);
  slotnames[0] = "charge";
  slotnames[1] = "current";
  MsgListAdd(object, "CURRENT", rxnpool_CURRENT, slotnames, 2);
  slotnames[0] = "A";
  MsgListAdd(object, "RXN0MOLES", rxnpool_RXN0MOLES, slotnames, 1);
  slotnames[0] = "vol";
  slotnames[1] = "SAin";
  slotnames[2] = "SAout";
  MsgListAdd(object, "VOLUME", rxnpool_VOLUME, slotnames, 3);
  object->description = "Concentration pool. 1-D or 2-D. Interacts with reactions, diffusion, pumps,\ncurrent, CICR or IICR.  Accumulates fluxes into A and B to solve d[C]/dt = A - B*[C]\nInput in Concentration for RXN0, RXN1 and RXN2 (0th, 1st and 2nd order reactions.\nInput for CURRENT in any units, set Iunits=1 for SI; Iunits=1e-12 for nA, msec\nUses physiologist convention: negative current is inward (increases Ca conc. \nInput from diffusion,CICR flux, M-M rxn or pump (molecules): RXN0MOLES \nInput in moles for RXN2MOLES (second order reaction, in units of moleculess:\nunits = 1 to use moles, 1e-3 for mmoles, 1e-6 for umoles\nDunits = 1 for meters, 1e-4 for cm, etc\nIf type = 0, converts molecule inputs to concentration and then integrates.\nIf type = 1, converts concentration inputs to molecules and then integrates.\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object cicr */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "cicr";
  tobject.type = "IP3R_type";
  tobject.size = sizeof(struct IP3R_type);
  { extern int cicrpool(); tobject.function = cicrpool; HashFunc("cicrpool", cicrpool, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("gate"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("cicr");
  object->defaults->object = object;
  object->defaults->name = CopyString("cicr");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Aug 99";
  { extern int cicrpool(); AddActionToObject(object, "INIT", cicrpool, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'cicr'\n")); HashFunc("cicrpool", cicrpool, "int"); }
  { extern int cicrpool(); AddActionToObject(object, "PROCESS", cicrpool, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'cicr'\n")); HashFunc("cicrpool", cicrpool, "int"); }
  { extern int cicrpool(); AddActionToObject(object, "RESET", cicrpool, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'cicr'\n")); HashFunc("cicrpool", cicrpool, "int"); }
  { extern int cicrpool(); AddActionToObject(object, "CHECK", cicrpool, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'cicr'\n")); HashFunc("cicrpool", cicrpool, "int"); }
  slotnames[0] = "kbalpha";
  slotnames[1] = "conc";
  MsgListAdd(object, "ASTATE", cicr_ASTATE, slotnames, 2);
  slotnames[0] = "kbbeta";
  slotnames[1] = "conc";
  MsgListAdd(object, "BSTATE", cicr_BSTATE, slotnames, 2);
  slotnames[0] = "kbgamma";
  slotnames[1] = "conc";
  MsgListAdd(object, "GSTATE", cicr_GSTATE, slotnames, 2);
  slotnames[0] = "ip3conc";
  MsgListAdd(object, "IP3", cicr_IP3, slotnames, 1);
  slotnames[0] = "caconc";
  MsgListAdd(object, "CALCIUM", cicr_CALCIUM, slotnames, 1);
  slotnames[0] = "xconc";
  MsgListAdd(object, "CONSERVE", cicr_CONSERVE, slotnames, 1);
  object->description = " computes state variables associated with ip3 receptor channel.\n Needs to send messages to cicrflux, which computes flux between two \n pools connected by the channel (i.e. cytosol and ER). See script for usage.\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object cicrflux */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "cicrflux";
  tobject.type = "flux_type";
  tobject.size = sizeof(struct flux_type);
  { extern int cicrflux(); tobject.function = cicrflux; HashFunc("cicrflux", cicrflux, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("channel"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("cicrflux");
  object->defaults->object = object;
  object->defaults->name = CopyString("cicrflux");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99; GMU Mar 02";
  { extern int cicrflux(); AddActionToObject(object, "INIT", cicrflux, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'cicrflux'\n")); HashFunc("cicrflux", cicrflux, "int"); }
  { extern int cicrflux(); AddActionToObject(object, "PROCESS", cicrflux, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'cicrflux'\n")); HashFunc("cicrflux", cicrflux, "int"); }
  { extern int cicrflux(); AddActionToObject(object, "RESET", cicrflux, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'cicrflux'\n")); HashFunc("cicrflux", cicrflux, "int"); }
  { extern int cicrflux(); AddActionToObject(object, "CHECK", cicrflux, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'cicrflux'\n")); HashFunc("cicrflux", cicrflux, "int"); }
  slotnames[0] = "conc1";
  MsgListAdd(object, "CONC1", cicrflux_CONC1, slotnames, 1);
  slotnames[0] = "conc2";
  MsgListAdd(object, "CONC2", cicrflux_CONC2, slotnames, 1);
  slotnames[0] = "ip3r";
  MsgListAdd(object, "IP3R", cicrflux_IP3R, slotnames, 1);
  object->description = "Computes flow through ip3 receptor channel, between ER and cytosol rxnpools.\nRequires inputs from two rxnpools (units must be concentration) and the ip3r object \nreturn messages as follows: \naddmsg from cicrflux object to pool 1 RXN0MOLES deltaflux2 \naddmsg from cicrflux object to pool 2 RXN0MOLES deltaflux1 \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object enzyme */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "enzyme";
  tobject.type = "enzyme_type";
  tobject.size = sizeof(struct enzyme_type);
  { extern int enzyme(); tobject.function = enzyme; HashFunc("enzyme", enzyme, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("enzyme");
  object->defaults->object = object;
  object->defaults->name = CopyString("enzyme");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99; revised GMU Mar 02";
  { extern int enzyme(); AddActionToObject(object, "INIT", enzyme, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'enzyme'\n")); HashFunc("enzyme", enzyme, "int"); }
  { extern int enzyme(); AddActionToObject(object, "PROCESS", enzyme, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'enzyme'\n")); HashFunc("enzyme", enzyme, "int"); }
  { extern int enzyme(); AddActionToObject(object, "RESET", enzyme, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'enzyme'\n")); HashFunc("enzyme", enzyme, "int"); }
  { extern int enzyme(); AddActionToObject(object, "SET", enzyme, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'enzyme'\n")); HashFunc("enzyme", enzyme, "int"); }
  { extern int enzyme(); AddActionToObject(object, "CHECK", enzyme, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'enzyme'\n")); HashFunc("enzyme", enzyme, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "ENZ", enzyme_ENZ, slotnames, 1);
  slotnames[0] = "feedback";
  slotnames[1] = "rateconst";
  MsgListAdd(object, "FEEDBACK", enzyme_FEEDBACK, slotnames, 2);
  slotnames[0] = "conc";
  MsgListAdd(object, "SUBSTRATE", enzyme_SUBSTRATE, slotnames, 1);
  slotnames[0] = "mrho";
  slotnames[1] = "vol";
  slotnames[2] = "SAin";
  slotnames[3] = "SAout";
  MsgListAdd(object, "RHODOPSIN", enzyme_RHODOPSIN, slotnames, 4);
  slotnames[0] = "vol";
  slotnames[1] = "SAin";
  slotnames[2] = "SAout";
  MsgListAdd(object, "VOLUME", enzyme_VOLUME, slotnames, 3);
  object->description = "Michaelis-Menton enzyme kinetics. Vmax = kcat; Km = (Kcat+kb)/kf\n			kf	  kcat\n[Enzyme] + [Substrate] <-> [E-S] -> [Enzyme] + [Product]\n			 kb\nComputes dprod/dt =  kcat*E-Scomplex\nSolves for E-Scomplex concentration from:\ndES/dt = -(kcat+kb)*ES + kf*E*S\nkcat and kb units are per time. Kf units must be\nconsistent with substrate units.\ndProduct/dt units are quantity (not concentration)\naddmsg from enz object to rxnpool RXN0MOLES product.\nFeedback allows decrease in rate constants as function of \nconcentration of feedback substance.  Feedback message requires \nfeedback conc and rate const: 0 for kf, 1 for kb, 2 for kcat.\nhyperbolic form: fb propto 1 / conc(feedback substance) for neg feedback \n                 fb propto conc (feeback substance) for positive feedback. \nSigmoid form: fb propto conc/conc + halfmax (for inc feedback) \n		fb propto halfmax / conc + halfmax (for dec feedback) \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object mmenz */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "mmenz";
  tobject.type = "mmenz_type";
  tobject.size = sizeof(struct mmenz_type);
  { extern int mmenz(); tobject.function = mmenz; HashFunc("mmenz", mmenz, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("mmenz");
  object->defaults->object = object;
  object->defaults->name = CopyString("mmenz");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99";
  { extern int mmenz(); AddActionToObject(object, "INIT", mmenz, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'mmenz'\n")); HashFunc("mmenz", mmenz, "int"); }
  { extern int mmenz(); AddActionToObject(object, "PROCESS", mmenz, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'mmenz'\n")); HashFunc("mmenz", mmenz, "int"); }
  { extern int mmenz(); AddActionToObject(object, "RESET", mmenz, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'mmenz'\n")); HashFunc("mmenz", mmenz, "int"); }
  { extern int mmenz(); AddActionToObject(object, "SET", mmenz, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'mmenz'\n")); HashFunc("mmenz", mmenz, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "ENZ", mmenz_ENZ, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "FEEDBACK", mmenz_FEEDBACK, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "SUBSTRATE", mmenz_SUBSTRATE, slotnames, 1);
  object->description = "Simplified Michaelis-Menton enzyme kinetics.  Can assume constant substrate, \n or can compute effect of substrate on reaction rate.  Vmax = kf2;  Km = (Kf2+kb1)/kf1\n			kf1	  kf2\n[Enzyme] + [Substrate] <-> [E-S] -> [Enzyme] + [Product]\n			 kb1\nUnder assumptions of M-M dynamics, Solves   dprod/dt =  kf1*Enzyme*Subs/(Subs + Km\nVmax units are mmol product per (mg or mmol or mM enzyme) per msec.  Must be consistent \nwith Enzyme units (input from enzyme rxnpool.  Substrate and Km must have units of Conc.\nProduct (mmol units) = Vmax * E * S/ (S+Km).  addmsg from enz object to rxnpool RXN0MOLES product.\nFor more accurate MM kinetics, use reaction objects OR Bhalla enz object in kinetikit.\nFeedback allows decrease in enzyme activity as function of concentration of feedback substance \nhyperbolic form: fb propto 1 / conc(feedback substance) for neg feedback \n                 fb propto conc (feeback substance) for positive feedback. \nSigmoid form: fb propto conc/conc + halfmax (for neg feedback) \n		fb propto halfmax / conc + halfmax (for pos feedback) \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object feedback */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "feedback";
  tobject.type = "feedback_type";
  tobject.size = sizeof(struct feedback_type);
  { extern int feedback(); tobject.function = feedback; HashFunc("feedback", feedback, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("feedback");
  object->defaults->object = object;
  object->defaults->name = CopyString("feedback");
  object->author = "A. Blackwell GMU feb 02";
  { extern int feedback(); AddActionToObject(object, "INIT", feedback, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'feedback'\n")); HashFunc("feedback", feedback, "int"); }
  { extern int feedback(); AddActionToObject(object, "PROCESS", feedback, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'feedback'\n")); HashFunc("feedback", feedback, "int"); }
  { extern int feedback(); AddActionToObject(object, "RESET", feedback, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'feedback'\n")); HashFunc("feedback", feedback, "int"); }
  { extern int feedback(); AddActionToObject(object, "SET", feedback, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'feedback'\n")); HashFunc("feedback", feedback, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "CONC", feedback_CONC, slotnames, 1);
  object->description = "An object for computing modification of enzyme rate constants\ncaused by feedback of some substance, e.g. calcium, sent as message.\nhyperbolic form: fb propto 1 / conc(feedback substance) for neg feedback \n                 fb propto conc (feeback substance) for positive feedback. \nSigmoid form: fb propto conc/conc + halfmax (for inc feedback) \n		fb propto halfmax / conc + halfmax (for dec feedback) \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object rhodopsin */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "rhodopsin";
  tobject.type = "rhodopsin_type";
  tobject.size = sizeof(struct rhodopsin_type);
  { extern int rhodopsin(); tobject.function = rhodopsin; HashFunc("rhodopsin", rhodopsin, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("rhodopsin");
  object->defaults->object = object;
  object->defaults->name = CopyString("rhodopsin");
  object->author = "A. Blackwell GMU March 02";
  { extern int rhodopsin(); AddActionToObject(object, "INIT", rhodopsin, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'rhodopsin'\n")); HashFunc("rhodopsin", rhodopsin, "int"); }
  { extern int rhodopsin(); AddActionToObject(object, "PROCESS", rhodopsin, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'rhodopsin'\n")); HashFunc("rhodopsin", rhodopsin, "int"); }
  { extern int rhodopsin(); AddActionToObject(object, "RESET", rhodopsin, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'rhodopsin'\n")); HashFunc("rhodopsin", rhodopsin, "int"); }
  { extern int rhodopsin(); AddActionToObject(object, "SET", rhodopsin, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'rhodopsin'\n")); HashFunc("rhodopsin", rhodopsin, "int"); }
  { extern int rhodopsin(); AddActionToObject(object, "CHECK", rhodopsin, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'rhodopsin'\n")); HashFunc("rhodopsin", rhodopsin, "int"); }
  slotnames[0] = "isomerization";
  MsgListAdd(object, "ISOM", rhodopsin_ISOM, slotnames, 1);
  slotnames[0] = "inactivation";
  MsgListAdd(object, "INACT", rhodopsin_INACT, slotnames, 1);
  object->description = "Stochastic activation of rhodopsin molecules by Poisson generator\nincludes stochastic inactivation by Rhodopsin Kinase or Arrestin\nKeeps track of: \n   Number of rhodopsin molecules per villi\n       which allows decrease effectiveness with additional villi\n   time of rhodopsin activation\n       which allows decrease effectiveness over time due to depletion\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object mmpump2 */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "mmpump2";
  tobject.type = "pump_type";
  tobject.size = sizeof(struct pump_type);
  { extern int mmpump2(); tobject.function = mmpump2; HashFunc("mmpump2", mmpump2, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("channel"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("mmpump2");
  object->defaults->object = object;
  object->defaults->name = CopyString("mmpump2");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99; GMU Mar 02";
  { extern int mmpump2(); AddActionToObject(object, "INIT", mmpump2, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'mmpump2'\n")); HashFunc("mmpump2", mmpump2, "int"); }
  { extern int mmpump2(); AddActionToObject(object, "PROCESS", mmpump2, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'mmpump2'\n")); HashFunc("mmpump2", mmpump2, "int"); }
  { extern int mmpump2(); AddActionToObject(object, "RESET", mmpump2, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'mmpump2'\n")); HashFunc("mmpump2", mmpump2, "int"); }
  { extern int mmpump2(); AddActionToObject(object, "SET", mmpump2, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'mmpump2'\n")); HashFunc("mmpump2", mmpump2, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "CONC", mmpump2_CONC, slotnames, 1);
  object->description = " A Michaelis-Menton type of formulation of pump.  Either SERCA ATPase pump or \n Na-Ca ATPase cytosolic pump. Requires input message in concentration units.\n addmsg from mmpump2 to cytosol-rxnpool RXN0MOLES moles_out.\n addmsg from mmpump2 to ER or extracellular rxnpool RXN0MOLES moles_in.\n Computes right hand side of dC/dt = (+/-)max_rate * Ca^n / (Ca^n + half_conc^n). \n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object ncx */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "ncx";
  tobject.type = "exchange_type";
  tobject.size = sizeof(struct exchange_type);
  { extern int ncx(); tobject.function = ncx; HashFunc("ncx", ncx, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("channel"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("ncx");
  object->defaults->object = object;
  object->defaults->name = CopyString("ncx");
  object->author = "A. Blackwell GMU Sept 02";
  { extern int ncx(); AddActionToObject(object, "INIT", ncx, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'ncx'\n")); HashFunc("ncx", ncx, "int"); }
  { extern int ncx(); AddActionToObject(object, "PROCESS", ncx, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'ncx'\n")); HashFunc("ncx", ncx, "int"); }
  { extern int ncx(); AddActionToObject(object, "RESET", ncx, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'ncx'\n")); HashFunc("ncx", ncx, "int"); }
  { extern int ncx(); AddActionToObject(object, "SET", ncx, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'ncx'\n")); HashFunc("ncx", ncx, "int"); }
  { extern int ncx(); AddActionToObject(object, "CHECK", ncx, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'ncx'\n")); HashFunc("ncx", ncx, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "CAINT", ncx_CAINT, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "CAEXT", ncx_CAEXT, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "NAINT", ncx_NAINT, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "NAEXT", ncx_NAEXT, slotnames, 1);
  slotnames[0] = "Vm";
  MsgListAdd(object, "VM", ncx_VM, slotnames, 1);
  object->description = " Sodium calcium exchanger equations.  if ncxtype = 0 uses Gall formula:\n I = K * Cai^Hill/(Cai^Hill + Kmca^H) * (Vm - Vnaca) \n where Vnaca = RT/F(stoic*ln(Nao/Nai) - ln(Cao/Cai)) \n if ncxtype = 1 uses jafri formula: \n I = K* (Cao*Nai^stoich*exp(gamma*FV/RT) - Cai*Nao^stoich*exp((1-gamma*FV/RT))\n divided by (Kmna^stoich+nao^stoich)*(Kmca+cao)*(1+ksat*exp(1-gamma)FV/RT)) \n if Na_msg = 0 then Na concentrations are initialized, no Na messages passed\n   Requires calcium and sodium input message in concentration units.\n Requires Vm message, units indicated with Vunit field \n addmsg from ncx to rxnpool CURRENT valence current\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object reaction */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "reaction";
  tobject.type = "react_type";
  tobject.size = sizeof(struct react_type);
  { extern int reaction(); tobject.function = reaction; HashFunc("reaction", reaction, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("reaction");
  object->defaults->object = object;
  object->defaults->name = CopyString("reaction");
  object->author = "A. Blackwell ERIM Aug 96 ; revised GMU Dec 99";
  { extern int reaction(); AddActionToObject(object, "INIT", reaction, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'reaction'\n")); HashFunc("reaction", reaction, "int"); }
  { extern int reaction(); AddActionToObject(object, "PROCESS", reaction, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'reaction'\n")); HashFunc("reaction", reaction, "int"); }
  { extern int reaction(); AddActionToObject(object, "RESET", reaction, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'reaction'\n")); HashFunc("reaction", reaction, "int"); }
  { extern int reaction(); AddActionToObject(object, "SET", reaction, 0) ? 0 : (Error(), printf("adding action 'SET' to object 'reaction'\n")); HashFunc("reaction", reaction, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "SUBSTRATE", reaction_SUBSTRATE, slotnames, 1);
  slotnames[0] = "conc";
  MsgListAdd(object, "PRODUCT", reaction_PRODUCT, slotnames, 1);
  object->description = "An element to compute kf*substrates and kb*products for nth order reactions.\ndC/dt = kf * substrate * substrate * ... - kb*product* ...\nFor each substrate and product in reaction, a message must be received (n times \nif n moles required). Input and rate constants both must be in units of concentration,\n(use RXN2 to send message to rxnpool) or both must be in units of moles (use RXN2MOLE).\naddmsg from reaction to substrate rxnpool RXN2 (or RXN2MOLES) kbprod kfsubs\naddmsg from reaction to product rxnpool RXN2 (or RXN2MOLES) kfsubs kbprod	\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object conservepool */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "conservepool";
  tobject.type = "conserve_type";
  tobject.size = sizeof(struct conserve_type);
  { extern int conservepool(); tobject.function = conservepool; HashFunc("conservepool", conservepool, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  ObjectAddClass(&tobject,ClassID("concentration"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("conservepool");
  object->defaults->object = object;
  object->defaults->name = CopyString("conservepool");
  object->author = "A. Blackwell ERIM Aug 96; revised GMU Dec 99; GMU Mar 02";
  { extern int conservepool(); AddActionToObject(object, "INIT", conservepool, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'conservepool'\n")); HashFunc("conservepool", conservepool, "int"); }
  { extern int conservepool(); AddActionToObject(object, "PROCESS", conservepool, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'conservepool'\n")); HashFunc("conservepool", conservepool, "int"); }
  { extern int conservepool(); AddActionToObject(object, "RESET", conservepool, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'conservepool'\n")); HashFunc("conservepool", conservepool, "int"); }
  { extern int conservepool(); AddActionToObject(object, "CHECK", conservepool, 0) ? 0 : (Error(), printf("adding action 'CHECK' to object 'conservepool'\n")); HashFunc("conservepool", conservepool, "int"); }
  slotnames[0] = "conc";
  MsgListAdd(object, "CONC", conservepool_CONC, slotnames, 1);
  slotnames[0] = "moles";
  MsgListAdd(object, "MOLES", conservepool_MOLES, slotnames, 1);
  slotnames[0] = "volume";
  MsgListAdd(object, "VOLUME", conservepool_VOLUME, slotnames, 1);
  object->description = "Implementation of pool for conserved molecules. If type = 0, computes\nconcentration of molecule which is the total conc less the conc in all\nother states. If type = 1, computes quantity of molecule which is the \ntotal quantity less the quantity in all other states.  Then computes\nconcentration by dividing by volume.\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

  /* Definition of object wgtavg */
  BZERO(&tobject,sizeof(GenesisObject));
  tobject.name = "wgtavg";
  tobject.type = "wgtavg_type";
  tobject.size = sizeof(struct wgtavg_type);
  { extern int wgtavg(); tobject.function = wgtavg; HashFunc("wgtavg", wgtavg, "int"); }
  ObjectAddClass(&tobject,ClassID("segment"),CLASS_PERMANENT);
  AddDefaultFieldList(&tobject);
  tobject.defaults = (Element*) calloc(1, tobject.size);
  AddObject(&tobject);
  object = GetObject("wgtavg");
  object->defaults->object = object;
  object->defaults->name = CopyString("wgtavg");
  object->author = "A. Blackwell GMU Oct 2011";
  { extern int wgtavg(); AddActionToObject(object, "INIT", wgtavg, 0) ? 0 : (Error(), printf("adding action 'INIT' to object 'wgtavg'\n")); HashFunc("wgtavg", wgtavg, "int"); }
  { extern int wgtavg(); AddActionToObject(object, "PROCESS", wgtavg, 0) ? 0 : (Error(), printf("adding action 'PROCESS' to object 'wgtavg'\n")); HashFunc("wgtavg", wgtavg, "int"); }
  { extern int wgtavg(); AddActionToObject(object, "RESET", wgtavg, 0) ? 0 : (Error(), printf("adding action 'RESET' to object 'wgtavg'\n")); HashFunc("wgtavg", wgtavg, "int"); }
  slotnames[0] = "Value";
  slotnames[1] = "wgt";
  MsgListAdd(object, "ValueWgt", wgtavg_ValueWgt, slotnames, 2);
  object->description = "An element to weighted average.  Divides total weighted value by total weights.\nCan calculated the mean calcium from unequal size calcium shells by using volume as weight.\n";
  FieldListMakePermanent(object);
  MsgListMakePermanent(object);

/* Script variables */

} /* STARTUP_chem */
